// Referred to http://ww1.microchip.com/downloads/en/appnotes/91054c.pdf for joystick HID descriptor
// Another good example was found at https://learn.adafruit.com/custom-hid-devices-in-circuitpython/report-descriptors

// PIC32MZ_USB_HID_joystick.c - A USB HID joystick implementation for the PIC32MZ
// by Aidan Mocke 2022-07-25

// Setup for PIC32MZ EF device for 100 MHz operation - Device Configuration Bits.
#pragma config DEBUG =      OFF
#pragma config JTAGEN =     OFF
#pragma config ICESEL =     ICS_PGx1
#pragma config TRCEN =      OFF
#pragma config BOOTISA =    MIPS32
#pragma config FECCCON =    OFF_UNLOCKED
#pragma config FSLEEP =     OFF
#pragma config DBGPER =     PG_ALL
#pragma config SMCLR =      MCLR_NORM
#pragma config SOSCGAIN =   GAIN_2X
#pragma config SOSCBOOST =  ON
#pragma config POSCGAIN =   GAIN_2X
#pragma config POSCBOOST =  ON
#pragma config EJTAGBEN =   NORMAL
#pragma config CP =         OFF

/*** DEVCFG1 ***/

#pragma config FNOSC =      SPLL
#pragma config DMTINTV =    WIN_127_128
#pragma config FSOSCEN =    OFF
#pragma config IESO =       OFF
#pragma config POSCMOD =    EC
#pragma config OSCIOFNC =   OFF
#pragma config FCKSM =      CSECME
#pragma config WDTPS =      PS1048576
#pragma config WDTSPGM =    STOP
#pragma config FWDTEN =     OFF
#pragma config WINDIS =     NORMAL
#pragma config FWDTWINSZ =  WINSZ_25
#pragma config DMTCNT =     DMT31
#pragma config FDMTEN =     OFF
/*** DEVCFG2 ***/

#pragma config FPLLIDIV =   DIV_3
#pragma config FPLLRNG =    RANGE_5_10_MHZ
#pragma config FPLLICLK =   PLL_POSC
#pragma config FPLLMULT =   MUL_50
#pragma config FPLLODIV =   DIV_4
#pragma config UPLLFSEL =   FREQ_24MHZ
/*** DEVCFG3 ***/

#pragma config USERID =     0xffff
#pragma config FMIIEN =     ON
#pragma config FETHIO =     ON
#pragma config PGL1WAY =    OFF
#pragma config PMDL1WAY =   OFF
#pragma config IOL1WAY =    OFF
#pragma config FUSBIDIO =   ON

#define SYS_FREQ (100000000L) // 100MHz should be more than enough for a joystick

#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

unsigned char USB_device_descriptor[18] = {
	0x12, 		// Descriptor length - Always 18 or 0x12
	0x01, 		// Descriptor type - Device
	0x00,0x02, 	// bcdUSB (USB version supported) - 2.0
	0x00, 		// Device class - 0 = get from configuration
	0x00, 		// Device sub-class - 0 = get from configuration
	0x00, 		// Device Protocol - 0 = get from configuration
	0x40, 		// Maximum packet size - Always 64 or 0x40 for High Speed USB on endpoint 0
	0xD8,0x04, 	// Vendor ID - 0x04D8 = Microchip Vendor ID
	0x55,0x00, 	// Product ID - 0x0055 but could be anything for our home usage
	0x11,0x01, 	// Device USB version - mine uses USB 1.11
	0x01, 		// Manufacturer string index number - when host asks for it
	0x02, 		// Product string index number
	0x00, 		// Serial number string index number - 0 = does not exist, tells host not to ask for it
	0x01        // Number of configurations for this device - 1 for us
};

unsigned char USB_config_descriptor[41] = {
    // BASIC DESCRIPTOR BEGINS HERE - this will be requested first
	0x09,		// Descriptor length (for basic descriptor - without config information)
	0x02,		// Descriptor type - 2 = Configuration
	0x29, 0x00,	// Total length of the descriptor, including config information - size is 0x0029 or 41 bytes
	0x1,		// Number of interfaces - 1 for our project
	0x1,		// Configuration number to start with - 1 again for us 
	0x0,		// Configuration string index - does not exist, so host won't ask for it
    0x80,       // bmAttributes - bit 7 set (0x80) means bus powered, 0xC0 is for self powered
    0x32,       // Max power used by device. Actual current is value x 2mA, so I have 2x50 = 100mA. Set to 0 for self powered.    
    // FULL DESCRIPTOR CONTINUES HERE - when the config descriptor is requested the second time this will be sent too
    // INTERFACE 0 - HID Class
    0x09,       // bLength of interface descriptor
    0x04,       // bDescriptortype - 0x04 is an INTERFACE
    0x00,       // bInterfaceNumber - This is the first interface
    0x00,       // bAlternateSetting - not used
    0x02,       // bNumEndpoints - We will use 1 endpoint for sending data and one for receiving
    0x03,       // bInterfaceClass - 0x03 means HID
    0x00,       // bInterfaceSubclass - not important here
    0x00,       // bInterfaceProtocol - not important here
    0x00,       // iInterface - index of string describing this interface, as always 0 means not used
    // HID DESCRIPTOR
    0x09,       // bLength
    0x21,       // bDescriptorType
    0x11,0x01,  // bcdHID - Supports USB 1.11
    0x00,       // bCountryCode
    0x01,       // bNummDescriptors - just the one for us
    0x22,       // bDescriptorSubtype - 0x22 is HID Report
    0x2D, 0x00, // wDescriptorLength - Our descriptor is 0x004A bytes long (45 bytes)
    // FIRST ENDPOINT DESCRIPTOR - IN TO host FROM device
    0x07,       // bLength
    0x05,       // bDescriptorType - 0x05 means ENDPOINT
    0x81,       // bEndpointAdress - 0x81 comes from using endpoint 0x1 and ORing it with 0x80 to signify it sends data IN TO the host
    0x03,       // bmAttributes - 0x03 means interrupt driven
    0x40,       // Max Packet Size (LSB) - 64 bytes at a time
    0x00,       // Max Packet Size (MSB)
    0x01,       // Polling interval - doesn't apply for us, we'll send updates when necessary    
    // SECOND ENDPOINT DESCRIPTOR - OUT FROM host IN TO device
    0x07,        // bLength
    0x05,        // bDescriptorType - 0x05 means ENDPOINT
    0x01,        // bEndpointAdress - 0x01 (NO ORing with 0x80 because we are receiving data FROM the host)
    0x03,        // bmAttributes - 0x03 means interrupt driven
    0x40,        // Max Packet Size (LSB) - 64 bytes at a time
    0x00,        // Max Packet Size (MSB)
    0x01,        // Polling interval - doesn't apply for us, we'll send updates when necessary
};

// Have separated out this portion of the config descriptor because it is apparently requested separately sometime (?) - it is the same as the one contained in USB_config_descriptor
unsigned char HID_descriptor[9] = {
    0x09,       // bLength
    0x21,       // bDescriptorType
    0x11,0x01,  // bcdHID - Supports USB 1.11
    0x00,       // bCountryCode
    0x01,       // bNummDescriptors - just the one for us
    0x22,       // bDescriptorSubtype - 0x22 is HID Report
    0x2D, 0x00, // wDescriptorLength - Our descriptor is 0x004A bytes long (45 bytes)
};

typedef union _JOYSTICK_CONTROLS_TYPEDEF
{
    struct
    {
        struct
        {
            uint8_t button1:1;
            uint8_t button2:1;
            uint8_t button3:1;
            uint8_t button4:1;
            uint8_t button5:1;
            uint8_t button6:1;
            uint8_t button7:1;
            uint8_t button8:1;
            uint8_t button9:1;
            uint8_t button10:1;
            uint8_t button11:1;
            uint8_t button12:1;
            uint8_t button13:1;
            uint8_t button14:1;
            uint8_t button15:1;
            uint8_t button16:1;
        } buttons;
        struct
        {
            uint8_t X;
            uint8_t Y;
        } analog_stick;
    } members;
    uint8_t val[4];
} JOYSTICK_STATE;

uint8_t joystick_report_descriptor[45] = {
  0x05,0x01,        // UsagePage (Generic Desktop)
  0x09,0x05,        // Usage (Game Pad)
  0xA1,0x01,        // Collection (Application)
  0x15,0x00,        // LogicalMinimum (0)
  0x25,0x01,        // LogicalMaximum (1)
  0x35,0x00,        // PhysicalMinimum (0)
  0x45,0x01,        // PhysicalMaximum (1)
  0x75,0x01,        // ReportSize (1)
  0x95,0x10,        // ReportCount (8)
  0x05,0x09,        // UsagePage (Button)
  0x19,0x01,        // UsageMinimum (Button 1)
  0x29,0x10,        // UsageMaximum (Button 8)
  0x81,0x02,        // Input (Data,Var,Abs)
  0x05,0x01,        // UsagePage (Generic Desktop)
  0x26,0xFF,0x00,   // LogicalMaximum (255)
  0x46,0xFF,0x00,   // PhysicalMaximum (255)
  0x09,0x30,        // Usage (X)
  0x09,0x31,        // Usage (Y)
  0x75,0x08,        // ReportSize (8)
  0x95,0x02,        // ReportCount (4)
  0x81,0x02,        // Input (Data,Var,Abs)
  0xC0              // EndCollection
};


#define	STR_DESC(l)	(((((l))<<1) & 0xFF) | 0x0300)
// String descriptors. All have a common format: [byte] string_length, [byte] data_type (3 = string), UTF-16 encoded string, hence all the 0's between each 8-byte character

// Language description
unsigned char USB_LANGUAGE_STRING[] =  {0x4, 0x03, 0x09, 0x04}; // String length 4, data type 3 (string), 0x0409 = English

// Vendor description
unsigned short USB_VENDOR_STRING[] = { STR_DESC(12),
                            'A','i','d','a','n',' ','M','o','c','k','e',0};   

// Product description
unsigned short USB_PRODUCT_STRING[] = { STR_DESC(19),
                            'H','a','n','n','a','h','\'','s',' ','J','o','y','s','t','i','c','k',0};
 
// Serial number description
unsigned short USB_SERIAL_NUMBER_STRING[] = { STR_DESC(7),
                            '1','2','3','-','3','2','1',0};

unsigned char dummy_buffer[4] =
{
    0,0,0,0
};

JOYSTICK_STATE joystick_state, old_joystick_state;

#define MAX_USB_BUFFER_SIZE 512
#define USB_EP0_WAIT_TIMEOUT 1000000
#define VBUS_VALID 0x3

#define JOYSTICK_UP PORTFbits.RF4
#define JOYSTICK_LEFT PORTFbits.RF5
#define JOYSTICK_DOWN PORTDbits.RD9
#define JOYSTICK_RIGHT PORTDbits.RD10
#define JOYSTICK_BUTTON1 PORTFbits.RF0
#define JOYSTICK_BUTTON2 PORTDbits.RD5
#define JOYSTICK_BUTTON3 PORTDbits.RD4
#define JOYSTICK_BUTTON4 PORTDbits.RD3
#define JOYSTICK_BUTTON5 PORTDbits.RD2
#define JOYSTICK_BUTTON6 PORTEbits.RE7
#define JOYSTICK_BUTTON7 PORTEbits.RE5
#define JOYSTICK_BUTTON8 PORTEbits.RE4

typedef struct
{
    int baud;
    short stop_bits;
    short parity_bits;
    short data_bits;
} USB_LINE_CODING;

typedef struct
{
    volatile unsigned short rx_num_bytes;
    volatile unsigned short tx_num_bytes;
    volatile unsigned char tx_buffer[MAX_USB_BUFFER_SIZE];
    volatile unsigned char rx_buffer[MAX_USB_BUFFER_SIZE];
} USB_ENDPOINT;

typedef struct
{
    volatile unsigned char bmRequestType;
    volatile unsigned char bmRequest;
    volatile unsigned short wValue;
    volatile unsigned short wIndex;
    volatile unsigned short wLength;
    unsigned char rx_buffer[8];
    volatile unsigned short signature;
} USB_TRANSACTION;

unsigned char USB_address = 0;
volatile char USB_set_address = 0;
volatile char USB_line_control_state = 0;

USB_ENDPOINT EP[2];
USB_TRANSACTION USB_transaction;
USB_LINE_CODING USB_line_coding;
unsigned char USB_KEYBOARD_REPORT[8];

volatile char USB_READY = 0;
volatile char USB_CONNECTED = 0;
volatile char USB_REPORT_SENT = 1;
volatile char USB_REPORT_RECEIVED = 0;

void USB_init_endpoints()
{
    USBE1CSR0bits.MODE = 0;     // EP1 is OUT (OUT from host = in to device = RX)
    
    // Clear all interrupt flags
	USBCSR0bits.EP0IF = 0;
    USBCSR0bits.EP1TXIF = 0;
    USBCSR1bits.EP1RXIF = 0;
	
    // Set the maximum transfer length for each endpoint
	// Configure endpoint 0 first.
    USBE0CSR0bits.TXMAXP = 64; // Set endpoint 0 buffer size to 64 bytes (always a maximum of 64 for EP0, have seen as low as 8 for some keyboards)

    // And next my custom endpoints
    USBE1CSR0bits.TXMAXP = 64;   // Endpoint 1 - Maximum TX payload size at one time (transactions bigger than this need to be split and sent piecemeal)
    USBE1CSR1bits.RXMAXP = 64;   // Endpoint 1 - Maximum RX payload size at one time

	// Specify which kinds of endpoint we will be using
    USBE1CSR2bits.PROTOCOL = 3; // Endpoint 1 - Interrupt mode
            
	// Enable DISNYET
    USBE1CSR1bits.PIDERR = 0; // Clear DISNYET to enable NYET replies

	// Set up buffer locations for endpoint 1
    USBCSR3bits.ENDPOINT = 1;
    USBFIFOA = 0x00A000A;    
    USBIENCSR0bits.CLRDT = 1;
    USBIENCSR3bits.SPEED = 3;
    USBIENCSR2bits.SPEED = 3;
    USBE1CSR3bits.TXFIFOSZ = 0x6;   // Transmit FIFO Size bits - 64 bytes
    USBE1CSR3bits.RXFIFOSZ = 0x6;   // Transmit FIFO Size bits - 64 bytes
	
	// Set maximum size for each packet before splitting occurs
    USBOTGbits.TXFIFOSZ = 0b0011; // 0b0011 = 64 bytes
    USBOTGbits.RXFIFOSZ = 0b0011; // 0b0011 = 64 bytes

    
	// Disable Isochronous mode for endpoints 1 and 2
    USBE1CSR0bits.ISO = 0;      // Isochronous TX Endpoint Disable bit (Device mode).

    // Set up endpoint interrupts5555
	// Initially clear all the interrupt enables (IE)
    USBCSR1 = 0;

    USBCSR1bits.EP0IE = 1;      // Endpoint 0 interrupt enable
    USBCSR1bits.EP1TXIE = 1;    // Endpoint 1 TX interrupt enable
    USBCSR2bits.EP1RXIE = 1;    // Endpoint 1 RX interrupt enable
    USBCSR2bits.RESETIE = 1;
    USBCSR2bits.RESUMEIE = 1;
    USBCSR2bits.SOFIE = 1;

	// Set current endpoint to EP1
    USBCSR3bits.ENDPOINT = 1;
}

void USB_init()
{
    USBCSR0bits.SOFTCONN = 0;   // Initially disable it while we go into setup

    USB_address = 0;			// We haven't been given an address yet    
    USBCSR0bits.FUNC = 0;       // Initially set the USB address to 0 until later when the host assigns us an address

    USBCSR2bits.RESETIE = 1;    // Enable the reset interrupt
    IEC4bits.USBIE = 1;         // Enable the USB interrupt
    USBCRCONbits.USBIE = 1;     // Enable USB module interrupt
    IFS4bits.USBIF = 0;         // Clear the USB interrupt flag.
    IPC33bits.USBIP = 7;        // USB Interrupt Priority 7
    IPC33bits.USBIS = 1;        // USB Interrupt Sub-Priority 1

    USBCSR0bits.HSEN = 1;       // Enable High Speed (480Mbps) USB mode
    
    USB_line_coding.baud = 9600;
    USB_line_coding.data_bits = 8;
    USB_line_coding.parity_bits = 0;
    USB_line_coding.stop_bits = 0;

    USB_CONNECTED = 0;          // We have not received any USB communication yet, set connected to false
}

int USB_EP0_Wait_TXRDY()
{
    int timeout;
    
    timeout = 0;
    
    while (USBE0CSR0bits.TXRDY)
    {
        timeout++;
        if (timeout > USB_EP0_WAIT_TIMEOUT) return 1;
    };
    
    return 0;
}

int USB_EP1_Wait_TXRDY()
{
    int timeout;
    
    timeout = 0;
    
    while (USBE1CSR0bits.TXPKTRDY)
    {
        timeout++;
        if (timeout > USB_EP0_WAIT_TIMEOUT) 
        {
            return 1;
        }
    };
    
    return 0;
}

// Send EP[0].tx_num_bytes from EP[0].tx_buffer on endpoint 0
void USB_send_buf_EP0(void *buffer, int actual_length, int max_length)
{
    int cnt;
    unsigned char *FIFO_buffer;
    unsigned char *send_buffer = buffer;
    unsigned char b;
    
    // Sometimes EP0 seems to get requests for data longer than I actually have or, more often and in spec, it only requests a piece of the information at a time
    if (max_length < actual_length)
        actual_length = max_length;
        
    FIFO_buffer = (unsigned char *)&USBFIFO0;
    
    cnt = 0;
    
    if (USB_EP0_Wait_TXRDY()) return;

    while (cnt < actual_length)
    {
        *FIFO_buffer = *send_buffer++; // Send the bytes
        
        cnt++;
        // Have we sent 64 bytes?
        if ((cnt > 0) && (cnt % 64 == 0))
        {
            // Set TXRDY and wait for it to be cleared before sending any more bytes
            USBE0CSR0bits.TXRDY = 1;            
            if (USB_EP0_Wait_TXRDY()) return;
        }
    }
    
//    if ((cnt % 64) != 0) 
    USBE0CSR0bits.TXRDY = 1; 
}

void USB_send_buf_EP1(unsigned char *buffer, int length)
{
    int cnt;
    unsigned char *FIFO_buffer;
        
    USBE1CSR0bits.MODE = 1;     // EP1 is TX
    USBE1TXAbits.TXFADDR = USB_address;

    FIFO_buffer = (unsigned char *)&USBFIFO1;
    
    if (USB_EP1_Wait_TXRDY()) return;
    
    cnt = 0;
    
    while (cnt < length)
    {
        *FIFO_buffer = buffer[cnt]; // Send the bytes
        cnt++;
        // Have we sent 8 bytes?
        if ((cnt > 0) && (cnt % 64 == 0))
        {
            // Set TXRDY and wait for it to be cleared before sending any more bytes
            USBE1CSR0bits.TXPKTRDY = 1;            
            if (USB_EP1_Wait_TXRDY()) return;
        }
    }
    
    if ((cnt % 64) != 0) USBE1CSR0bits.TXPKTRDY = 1; 
    
    if (USB_EP1_Wait_TXRDY()) return;
}

void USB_send_joystick_report()
{
    USB_send_buf_EP1(joystick_state.val, sizeof(joystick_state.val));
}

void USB_receive_EP0(int length)
{
    int cnt;
    unsigned char *FIFO_buffer;
    
    // Store number of bytes received
    EP[0].rx_num_bytes = USBE0CSR2bits.RXCNT;
                
    // Get 8-bit pointer to USB FIFO for endpoint 0
    FIFO_buffer = (unsigned char *)&USBFIFO0;
    
    for(cnt = 0; cnt < length; cnt++)
    {
        // Read in one byte at a time
        EP[0].rx_buffer[cnt] = *(FIFO_buffer + (cnt & 3));
    }
        
    USBE0CSR0bits.RXRDYC = 1;
}

int USB_receive_EP1()
{
    unsigned char *FIFO_buffer;
    int cnt;
    int bytes_received;
    
    USBE1CSR0bits.MODE = 0;     // EP1 is RX

    bytes_received = USBE1CSR2bits.RXCNT;
    
    FIFO_buffer = (unsigned char *)&USBFIFO1;
    
    for(cnt = 0; cnt < bytes_received; cnt++)
    {
        EP[1].rx_buffer[cnt] = *(FIFO_buffer + (cnt & 3)); // Store the received bytes in array ep0data[].
    }
    
    USBE1CSR1bits.RXPKTRDY = 0;
    
    return bytes_received;
}

void USB_EP0_handle_transaction()
{
    int cnt;
    unsigned char temp_data[7];
    
    // We're not going to bother with whether bmRequestType is IN or OUT for the most part
    switch (USB_transaction.bmRequest)
    {
        case 0xC: // Sync frame
        {
            USBE0CSR0bits.STALL = 1;
            break;
            
        }
        case 0x0: 
        {
            if (USB_transaction.bmRequestType == 0x80) // Get status
                USB_send_buf_EP0(USB_device_descriptor, 0, 0);
            if (USB_transaction.bmRequestType == 0x00) // Select function
                USB_send_buf_EP0(USB_device_descriptor, 0, 0);
            break;            
        }
        
        case 0x1: // Get report via control endpoint
        {
            // Required to support this, just send 4 0's to make it happy
            if (USB_transaction.signature == 0xA101)
                USB_send_buf_EP0(dummy_buffer, sizeof(dummy_buffer), USB_transaction.wLength);
            
            break;
        }
        
        case 0x2: // Get idle
        {
            break;
        }
        
        case 0x3: // Get protocol
        {
            break;
        }
        
        case 0xB: // Set protocol
        {
            break;
        }
        
        case 0xA: // Set idle
        {
            USB_READY = 1;
            break;
        }
        
        case 0x5: // Set USB address
        {
            USBE0CSR0bits.RXRDYC = 1;
            USB_address = EP[0].rx_buffer[2] & 0x7F; // Can be 0 to 127, discard MSB
            USB_set_address = 1; // Tell our system we need to set the address now
            break;
        }

        case 0x6: // Get descriptor
        {
            switch (USB_transaction.wValue >> 8)
            {
                case 0x1: // Get device descriptor
                {
                    USB_send_buf_EP0(USB_device_descriptor, sizeof(USB_device_descriptor), USB_transaction.wLength);
                    break;
                }
                
                case 0x2: // Get configuration descriptor
                {
                    USB_send_buf_EP0(USB_config_descriptor, sizeof(USB_config_descriptor), USB_transaction.wLength);
                    break;
                }
                
                case 0x3: // Get string descriptors
                {
                    switch (USB_transaction.wValue & 0xFF)
                    {
                        case 0x0: // String 0 - Language ID
                        {
                            USB_send_buf_EP0(USB_LANGUAGE_STRING, sizeof(USB_LANGUAGE_STRING), USB_transaction.wLength);
                            break;
                        }
                        case 0x1: // String 1 - Vendor
                        {
                            USB_send_buf_EP0(USB_VENDOR_STRING, sizeof(USB_VENDOR_STRING), USB_transaction.wLength);
                            break;
                        }
                        case 0x2: // String 2 - Product name
                        {
                            USB_send_buf_EP0(USB_PRODUCT_STRING, sizeof(USB_PRODUCT_STRING), USB_transaction.wLength);
                            break;
                        }
                        case 0x3: // String 3 - Serial number
                        {
                            USB_send_buf_EP0(USB_SERIAL_NUMBER_STRING, sizeof(USB_SERIAL_NUMBER_STRING), USB_transaction.wLength);
                            break;
                        }
                        default: 
                        {
                            // Unused, shouldn't be asked for, but send a null string if it asks
                            USB_send_buf_EP0(USB_SERIAL_NUMBER_STRING, 0, USB_transaction.wLength);
                            break;
                        }
                    }
                    
                    case 0x21: // HID descriptor request
                    {
                        USB_send_buf_EP0(HID_descriptor, sizeof(HID_descriptor), USB_transaction.wLength);
                        break;
                    }
                    
                    case 0x22: // HID report descriptor request
                    {
                        {
                            USB_send_buf_EP0(joystick_report_descriptor, sizeof(joystick_report_descriptor), USB_transaction.wLength);
                        }
                        break;                        
                    }
                }
                default: 
                {
                    break;
                }
            }
            break;
        }
        
        case 0x9: // Set report
        {            
            USBE0CSR0bits.RXRDYC = 1; // ACK
            break;
        }
        
        case 0x20:
        {
            if (USB_transaction.bmRequestType == 0x21)
            {
                // Set line coding
                // Read (wLength) bytes from USB Endpoint 0
                USB_receive_EP0(USB_transaction.wLength);
                
                USB_line_coding.baud = ((int)EP[0].rx_buffer[3] << 24) | ((int)EP[0].rx_buffer[2] << 16) | ((int)EP[0].rx_buffer[1] << 8) | EP[0].rx_buffer[0];
                USB_line_coding.parity_bits = EP[0].rx_buffer[4];
                USB_line_coding.stop_bits = EP[0].rx_buffer[5];
                USB_line_coding.data_bits = EP[0].rx_buffer[6];
            }
            break;          
        }
        
        case 0x21:
        {
            // Get line coding
            temp_data[0] = USB_line_coding.baud & 0xFF;
            temp_data[1] = (USB_line_coding.baud >> 8) & 0xFF;
            temp_data[2] = (USB_line_coding.baud >> 16) & 0xFF;
            temp_data[3] = USB_line_coding.baud >> 24;
            temp_data[4] = USB_line_coding.stop_bits;
            temp_data[5] = USB_line_coding.parity_bits;
            temp_data[6] = USB_line_coding.data_bits;
            USB_send_buf_EP0(temp_data, sizeof(temp_data), USB_transaction.wLength);
            break;
        }
                
        default: 
        {
            USBE0CSR0bits.STALL = 1;
            break;
        }
    }   
}

void __attribute__((vector(_USB_VECTOR), interrupt(ipl7srs), nomips16)) USB_handler()
{
    unsigned char RESETIF, EP0IF, EP1TXIF, EP1RXIF;
    unsigned int CSR0, CSR1, CSR2;
    unsigned char bytes_received;
    
    USB_CONNECTED = 1; // USB communication has successfully been established
    
    // Save registers to temporary variables because some of them reset when you read them
    CSR0 = USBCSR0;
    EP0IF = (CSR0 & (1<<16)) ? 1 : 0;
    EP1TXIF = (CSR0 & (1<<17)) ? 1 : 0;    
    CSR1 = USBCSR1;    
    EP1RXIF = (CSR1 & (1 << 1)) ? 1 : 0;    
    CSR2 = USBCSR2;    
    RESETIF = (CSR2 & (1 << 18)) ? 1 : 0;
    
    if(RESETIF)
    {
        // So when this is set, the host's USB bus has been reset and we actually need to set up all the USB endpoints now. This will always occur at least once, when the host initialises itself.
        USB_init_endpoints();        
        USBCSR2bits.RESETIF = 0;
    }
    
    if(EP0IF) // Has an endpoint 0 interrupt been triggered?
    {
        if (USB_set_address) // Do we need the set the USB address?
        {
            USBCSR0bits.FUNC = USB_address;
            USB_set_address = 0;
            USBE1TXAbits.TXFADDR = USB_address;
            USBE1RXAbits.RXFADDR = USB_address;
        }
        
        if(USBE0CSR0bits.RXRDY)
        {
            USB_receive_EP0(USBE0CSR2bits.RXCNT);
            
            USB_transaction.bmRequestType = EP[0].rx_buffer[0];
            USB_transaction.bmRequest = EP[0].rx_buffer[1];
            USB_transaction.signature = ((int)EP[0].rx_buffer[0] << 8) | EP[0].rx_buffer[1];
            USB_transaction.wValue = (int)(EP[0].rx_buffer[3] << 8) | EP[0].rx_buffer[2];
            USB_transaction.wIndex = (int)(EP[0].rx_buffer[5] << 8) | EP[0].rx_buffer[4];
            USB_transaction.wLength = (int)(EP[0].rx_buffer[7] << 8) | EP[0].rx_buffer[6];
            
            USB_EP0_handle_transaction();
            
            if (USB_transaction.wLength == 0) USBE0CSR0bits.DATAEND = 1; // End of Data Control bit (Device mode) 
        }
                
        if (USBE0CSR0bits.SETEND) 
        {
            USBE0CSR0bits.SETENDC = 1;
        }
        
        USBCSR0bits.EP0IF = 0;  // Clear the USB EndPoint 0 Interrupt Flag.
    }
    
    if(EP1TXIF)
    { 
        USBCSR0bits.EP1TXIF = 0;
        USBCSR3bits.ENDPOINT = 1;
        USBIENCSR0bits.MODE = 0;
    } 
        
    if (USBE1CSR1bits.RXPKTRDY || EP1RXIF)
    { 
        USBCSR1bits.EP1RXIF = 0;
        if (USBE1CSR2bits.RXCNT == 0)
        {
            USBE0CSR0bits.RXRDYC = 1;
        }
        else
        {
            // Receive HID report (does this ever happen with a joystick?)
            bytes_received = USB_receive_EP1();
        }
    } 

    if (USBCSR2bits.SOFIF)
    {
        USBCSR2bits.SOFIF = 0;
    }
    
    IFS4bits.USBIF = 0;
}

void USB_connect()
{
    USBCSR0bits.SOFTCONN = 1;
}

long set_performance_mode()
{   
	unsigned int cp0;
	
    // Unlock Sequence
    asm volatile("di"); // Disable all interrupts
    SYSKEY = 0xAA996655;
    SYSKEY = 0x556699AA;  

    // PB1DIV
    // Peripheral Bus 1 cannot be turned off, so there's no need to turn it on
    PB1DIVbits.PBDIV = 1; // Peripheral Bus 1 Clock Divisor Control (PBCLK1 is SYSCLK divided by 2)

    // PB2DIV
    PB2DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB2DIVbits.PBDIV = 1; // Peripheral Bus 2 Clock Divisor Control (PBCLK2 is SYSCLK divided by 2)

    // PB3DIV
    PB3DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB3DIVbits.PBDIV = 1; // Peripheral Bus 3 Clock Divisor Control (PBCLK3 is SYSCLK divided by 2)

    // PB4DIV
    PB4DIVbits.ON = 1; // Peripheral Bus 4 Output Clock Enable (Output clock is enabled)
    while (!PB4DIVbits.PBDIVRDY); // Wait until it is ready to write to
    PB4DIVbits.PBDIV = 0; // Peripheral Bus 4 Clock Divisor Control (PBCLK4 is SYSCLK divided by 1)

    // PB5DIV
    PB5DIVbits.ON = 1; // Peripheral Bus 5 Output Clock Enable (Output clock is enabled)
    PB5DIVbits.PBDIV = 1; // Peripheral Bus 5 Clock Divisor Control (PBCLK5 is SYSCLK divided by 2)

    // PB7DIV
    PB7DIVbits.ON = 1; // Peripheral Bus 7 Output Clock Enable (Output clock is enabled)
    PB7DIVbits.PBDIV = 0; // Peripheral Bus 7 Clock Divisor Control (PBCLK7 is SYSCLK divided by 1)

    // PB8DIV
    PB8DIVbits.ON = 1; // Peripheral Bus 8 Output Clock Enable (Output clock is enabled)
    PB8DIVbits.PBDIV = 1; // Peripheral Bus 8 Clock Divisor Control (PBCLK8 is SYSCLK divided by 2)

    // PRECON - Set up prefetch
    PRECONbits.PFMSECEN = 0; // Flash SEC Interrupt Enable (Do not generate an interrupt when the PFMSEC bit is set)
    PRECONbits.PREFEN = 0b11; // Predictive Prefetch Enable (Enable predictive prefetch for any address)
    PRECONbits.PFMWS = 0b010; // PFM Access Time Defined in Terms of SYSCLK Wait States (Two wait states)
    
    CFGCONbits.USBSSEN = 1;

    // Set up caching
    cp0 = _mfc0(16, 0);
    cp0 &= ~0x07;
    cp0 |= 0b011; // K0 = Cacheable, non-coherent, write-back, write allocate
    _mtc0(16, 0, cp0);  

   // Configure SRS Priority Selection
    PRISSbits.SS0 = 0;
    PRISSbits.PRI1SS = 1;
    PRISSbits.PRI2SS = 2;
    PRISSbits.PRI3SS = 3;
    PRISSbits.PRI4SS = 4;
    PRISSbits.PRI5SS = 5;
    PRISSbits.PRI6SS = 6;
    PRISSbits.PRI7SS = 7;
    
    INTCONbits.MVEC = 1; // Global interrupt enable

    // Lock Sequence
    SYSKEY = 0x33333333;
    
    asm volatile("ei"); // Enable all interrupts
}

void setup_ports()
{
    // Turn off all analog inputs, make everything digital
    ANSELB = 0;
    ANSELE = 0;
    ANSELG = 0;
    
    // Make all ports outputs initially
    TRISB = 0;
    TRISC = 0;
    TRISD = 0;
    TRISE = 0;
    TRISF = 0;
    TRISG = 0;
    
    // Set all port values low initially
    LATB = 0;
    LATC = 0;
    LATD = 0;
    LATE = 0;
    LATF = 0;
    LATG = 0;
    
    LATFbits.LATF3 = 1; // USB device mode
    CNPUFbits.CNPUF3 = 1; // Pull up to ensure device mode

    TRISDSET = 0b11000111110; // RD1 -> RD5, RD9, RD10
    TRISESET = 0b11110000; // RE4 -> RE7
    TRISFSET = 0b110001; // RF0, RF4, RF5
    TRISGSET = 0b11000000; // RG6 + RG7
}

void delay_us(unsigned int us)
{
    
    // Convert microseconds us into how many clock ticks it will take
    us *= SYS_FREQ / 1000000 / 2; // Core Timer updates every 2 ticks

    _CP0_SET_COUNT(0); // Set Core Timer count to 0

    while (us > _CP0_GET_COUNT()); // Wait until Core Timer count reaches the number we calculated earlier
}

void delay_ms(unsigned int ms)
{
    delay_us(ms * 1000);
}

void reset_joystick_state()
{
    int cnt;
    
    // Reset Buttons to 0 - not pressed
    joystick_state.val[0] = 0x00;
    joystick_state.val[1] = 0x00;

    // "Analog" sticks - in my case digital, so 0 is max one direction, 0x80 is centered and 0xFF is max other direction
    joystick_state.val[2] = 0x80; // X
    joystick_state.val[3] = 0x80; // Y
    
    // Reset the "old" values too
    for (cnt = 0; cnt < sizeof(joystick_state.val); cnt++)
        old_joystick_state.val[cnt] = joystick_state.val[cnt];
}

unsigned char update_joystick_state()
{
    int cnt;
    unsigned char ret_val = 0;
    
    // Check up key
    if ((JOYSTICK_UP == 0))
        joystick_state.members.analog_stick.Y = 0;
    else
    // If up isn't pressed, check down key
    if ((JOYSTICK_DOWN == 0))
        joystick_state.members.analog_stick.Y = 255;
    else // If neither up nor down are pressed, reset Y axis to mid-point of 0x80
        joystick_state.members.analog_stick.Y = 0x80;                

    // Check left key
    if ((JOYSTICK_LEFT == 0))
        joystick_state.members.analog_stick.X = 0;
    else
    // If left isn't pressed, check right key
    if ((JOYSTICK_RIGHT == 0))
        joystick_state.members.analog_stick.X = 255;
    else // If neither are pressed, reset X axis to mid-point of 0x80
        joystick_state.members.analog_stick.X = 0x80;

    // Check button 1
    if ((JOYSTICK_BUTTON1 == 0))
        joystick_state.members.buttons.button1 = 1;
    else
        joystick_state.members.buttons.button1 = 0;

    // Check button 2
    if ((JOYSTICK_BUTTON2 == 0))
        joystick_state.members.buttons.button2 = 1;
    else
        joystick_state.members.buttons.button2 = 0;

    // Check button 3
    if ((JOYSTICK_BUTTON3 == 0))
        joystick_state.members.buttons.button3 = 1;
    else
        joystick_state.members.buttons.button3 = 0;

    // Check button 4
    if ((JOYSTICK_BUTTON4 == 0))
        joystick_state.members.buttons.button4 = 1;
    else
        joystick_state.members.buttons.button4 = 0;

    // Check button 5
    if ((JOYSTICK_BUTTON5 == 0))
        joystick_state.members.buttons.button5 = 1;
    else
        joystick_state.members.buttons.button5 = 0;

    // Check button 6
    if ((JOYSTICK_BUTTON6 == 0))
        joystick_state.members.buttons.button6 = 1;
    else
        joystick_state.members.buttons.button6 = 0;

    // Check button 7
    if ((JOYSTICK_BUTTON7 == 0))
        joystick_state.members.buttons.button7 = 1;
    else
        joystick_state.members.buttons.button7 = 0;

    // Check button 8
    if ((JOYSTICK_BUTTON8 == 0))
        joystick_state.members.buttons.button8 = 1;
    else
        joystick_state.members.buttons.button8 = 0;    
    
    for (cnt = 0; cnt < sizeof(joystick_state.val); cnt++)
            if (old_joystick_state.val[cnt] != joystick_state.val[cnt]) ret_val = 1;
    
    // Store the current values as the "old" values for next time's comparison
    for (cnt = 0; cnt < sizeof(joystick_state.val); cnt++)
        old_joystick_state.val[cnt] = joystick_state.val[cnt];                               
    
    return ret_val;
}

void USB_wait_for_connection()
{
    while (!USB_CONNECTED)
    {
        if((USBOTGbits.VBUS == VBUS_VALID) && !USB_CONNECTED)
        {
            USB_connect();
            delay_ms(1);
        }
    }    
}

void main()
{
    set_performance_mode();
    setup_ports();
    
    // delay_ms(3000); // PIC32MZxxxxECxxx series startup delay required
    
    reset_joystick_state();

    // Initialise USB
    USB_init();    
    USB_wait_for_connection();
          
    while (1)
    {
        if (USB_READY)
        {
            // Only send a report if something has changed, i.e. a new button has been pressed 
            if (update_joystick_state() == 1)
                USB_send_joystick_report();
            
            delay_ms(1); // A tiny delay before polling again
        }
    }
}